# SilentTimer Action Figure communication library
# libstaf.py
#
# 09.08.2022
# Copyright (C) 2022 Balthasar Szczepański
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from enum import IntEnum, IntFlag, Enum
from typing import NamedTuple, Union
from datetime import datetime, timezone
import serial

class STAF:
    TIMEOUT = 0.02
    BUFFER_SIZE = 32

    class Event(IntFlag):
        NONE      = 0x00
        TICK      = 0x01
        BUTTON_A  = 0x02
        BUTTON_B  = 0x04
        BUTTON_C  = 0x08
        ALARM     = 0x10
        COUNTDOWN = 0x40

    class Command(IntEnum):
        GET_TIME        = ord('G')
        GET_MODE        = ord('H')
        GET_DISPLAY     = ord('I')
        GET_VALUE       = ord('J')
        GET_CALCULATION = ord('K')
        GET_RUN_MODE    = ord('L')
        GET_VERSION     = ord('M')
        GET_PINOUT      = ord('N')
        GET_EVENT       = ord('O')
        GET_ALARM       = ord('P')
        GET_DIGIT       = ord('Q')
        READ_MEMORY     = ord('R')
        SET_DISPLAY       = ord('g')
        FREE_DISPLAY      = ord('h')
        TRIGGER_EVENT     = ord('i')
        GOTO_MODE         = ord('j')
        SELECT_DIGIT      = ord('k')
        SET_DIGIT_VALUE   = ord('l')
        CONFIRM           = ord('m')
        SET_TIME          = ord('n')
        SET_ALARM         = ord('o')
        SET_DATE          = ord('p')
        SET_YEAR          = ord('q')
        SET_COUNTDOWN     = ord('r')
        SET_CALIBRATION   = ord('s')
        SET_DEBUG_ADDRESS = ord('t')
        SET_DEBUG_VALUE   = ord('u')
        CALCULATE         = ord('v')
        SET_VALUE         = ord('w')
        SET_RUN_MODE      = ord('x')
        SET_BUZZER_MODE   = ord('y')
        WRITE_MEMORY      = ord('z')

    class Mode(IntEnum):
        TIME        = 0x00
        ALARM       = 0x01
        DATE        = 0x02
        COUNTUP     = 0x03
        COUNTDOWN   = 0x04
        CALIBRATION = 0x05
        VERSION     = 0x06
        DEBUG       = 0x07
        RESET       = 0x08
        CALCULATOR  = 0x09
        SET_TIME          = 0x80
        SET_ALARM         = 0x81
        SET_DATE          = 0x82
        SET_YEAR          = 0x83
        SET_COUNTDOWN     = 0x84
        SET_CALIBRATION   = 0x85
        SET_DEBUG_ADDRESS = 0x86
        SET_DEBUG_VALUE   = 0x87
        SET_CALC_NUMBER   = 0x88
        SET_CALC_OPERATOR = 0x89

    BUZZER = ord('U')
    EndFrame = [
        ord('\r'),
        ord('\n')
    ]
    GetCommand = range(ord('G'), ord('Z')+1)
    SetCommand = range(ord('g'), ord('z')+1)

    class Register(IntEnum):
        RTCHSEC   = 0x00
        RTCSEC    = 0x01
        RTCMIN    = 0x02
        RTCHOUR   = 0x03
        RTCWKDAY  = 0x04
        RTCDATE   = 0x05
        RTCMTH    = 0x06
        RTCYEAR   = 0x07
        CONTROL   = 0x08
        OSCTRIM   = 0x09
        ALM0SEC   = 0x0C
        ALM0MIN   = 0x0D
        ALM0HOUR  = 0x0E
        ALM0WKDAY = 0x0F
        ALM0DATE  = 0x10
        ALM0MTH   = 0x11
        ALM1HSEC  = 0x12
        ALM1SEC   = 0x13
        ALM1MIN   = 0x14
        ALM1HOUR  = 0x15
        ALM1WKDAY = 0x16
        ALM1DATE  = 0x17
        PWRDNMIN  = 0x18
        PWRDNHOUR = 0x19
        PWRDNDATE = 0x1A
        PWRDNMTH  = 0x1B
        PWRUPMIN  = 0x1C
        PWRUPHOUR = 0x1D
        PWRUPDATE = 0x1E
        PWRUPMTH  = 0x1F
        # in SRAM by STAF software:
        ALARMFLAG = 0x20
        ALARMMIN  = 0x21
        ALARMHOUR = 0x22
        CDNFLAG   = 0x23
        CDNSEC    = 0x24
        CDNMIN    = 0x25
        YEARFLAG  = 0x26
        YEARLOW   = 0x27
        YEARHIGH  = 0x28

    ALARMFLAG = 0xA5

    class BuzzerMode(IntEnum):
        ENABLED  = 0x00
        DISABLED = 0x01
        FORCED   = 0x02

    class CalculationFlag(IntFlag):
        NEGATIVE   = 0b00000001
        DEC_POINT  = 0b00000010
        NEGATIVE2  = 0b00000100
        DEC_POINT2 = 0b00001000
        NEGATIVE1  = 0b00010000
        DEC_POINT1 = 0b00100000
        DIVIDE     = 0b11000000
        MULTIPLY   = 0b10000000
        SUBTRACT   = 0b01000000
        ADD        = 0b00000000
        OPERATOR   = 0b11000000

    class Calculation(IntEnum):
        ADD        = 0b00000000
        SUBTRACT   = 0b01000000
        MULTIPLY   = 0b10000000
        DIVIDE     = 0b11000000

    class Display(IntFlag):
        none = 0x00000000
        a0 = 0x00000001
        b0 = 0x00000002
        c0 = 0x00000004
        d0 = 0x00000008
        e0 = 0x00000010
        f0 = 0x00000020
        g0 = 0x00000040
        h0 = 0x00000080
        a1 = 0x00000100
        b1 = 0x00000200
        c1 = 0x00000400
        d1 = 0x00000800
        e1 = 0x00001000
        f1 = 0x00002000
        g1 = 0x00004000
        h1 = 0x00008000
        a2 = 0x00010000
        b2 = 0x00020000
        c2 = 0x00040000
        d2 = 0x00080000
        e2 = 0x00100000
        f2 = 0x00200000
        g2 = 0x00400000
        h2 = 0x00800000
        a3 = 0x01000000
        b3 = 0x02000000
        c3 = 0x04000000
        d3 = 0x08000000
        e3 = 0x10000000
        f3 = 0x20000000
        g3 = 0x40000000
        h3 = 0x80000000
        n0  = 0x00000001
        ne0 = 0x00000002
        se0 = 0x00000004
        s0  = 0x00000008
        sw0 = 0x00000010
        nw0 = 0x00000020
        mid0= 0x00000040
        n1  = 0x00000100
        ne1 = 0x00000200
        se1 = 0x00000400
        s1  = 0x00000800
        sw1 = 0x00001000
        nw1 = 0x00002000
        mid1= 0x00004000
        n2  = 0x00010000
        ne2 = 0x00020000
        se2 = 0x00040000
        s2  = 0x00080000
        sw2 = 0x00100000
        nw2 = 0x00200000
        mid2= 0x00400000
        n3  = 0x01000000
        ne3 = 0x02000000
        se3 = 0x04000000
        s3  = 0x08000000
        sw3 = 0x10000000
        nw3 = 0x20000000
        mid3= 0x40000000
        dot_top     = 0x00000080
        dot_bottom  = 0x00008000
        dot_alarm   = 0x80000000

    class HSEC(IntFlag):
        HSECONE0 = 0b00000001
        HSECONE1 = 0b00000010
        HSECONE2 = 0b00000100
        HSECONE3 = 0b00001000
        HSECTEN0 = 0b00010000
        HSECTEN1 = 0b00100000
        HSECTEN2 = 0b01000000
        HSECTEN3 = 0b10000000
        HSECONE  = 0b00001111
        HSECTEN  = 0b11110000
        HSEC     = 0b11111111

    class SEC(IntFlag):
        SECONE0 = 0b00000001
        SECONE1 = 0b00000010
        SECONE2 = 0b00000100
        SECONE3 = 0b00001000
        SECTEN0 = 0b00010000
        SECTEN1 = 0b00100000
        SECTEN2 = 0b01000000
        ST      = 0b10000000
        SECONE  = 0b00001111
        SECTEN  = 0b01110000
        SEC     = 0b01111111

    class MIN(IntFlag):
        MINONE0 = 0b00000001
        MINONE1 = 0b00000010
        MINONE2 = 0b00000100
        MINONE3 = 0b00001000
        MINTEN0 = 0b00010000
        MINTEN1 = 0b00100000
        MINTEN2 = 0b01000000
        MINONE  = 0b00001111
        MINTEN  = 0b01110000
        MIN     = 0b01111111

    class HOUR(IntFlag):
        HOURONE0  = 0b00000001
        HOURONE1  = 0b00000010
        HOURONE2  = 0b00000100
        HOURONE3  = 0b00001000
        HOURTEN0  = 0b00010000
        HOURTEN1  = 0b00100000
        AM_PM     = 0b00100000
        H12_14    = 0b01000000
        TRIMSIGN  = 0b10000000
        HOURONE   = 0b00001111
        HOURTEN12 = 0b00010000
        HOURTEN24 = 0b00110000
        HOUR12    = 0b00011111
        HOUR24    = 0b00111111

    class WKDAY(IntFlag):
        WKDAY0   = 0b00000001
        WKDAY1   = 0b00000010
        WKDAY2   = 0b00000100
        VBATEN   = 0b00001000
        PWRFAIL  = 0b00010000
        OSCRUN   = 0b00100000
        ALMxIF   = 0b00001000
        ALMnMSK0 = 0b00010000
        ALMnMSK1 = 0b00100000
        ALMnMSK2 = 0b01000000
        ALMnPIN  = 0b10000000
        WKDAY    = 0b00000111
        ALMnMSK  = 0b01110000

    class DATE(IntFlag):
        DATEONE0 = 0b00000001
        DATEONE1 = 0b00000010
        DATEONE2 = 0b00000100
        DATETEN0 = 0b00001000
        DATETEN1 = 0b00010000
        DATEONE  = 0b00001111
        DATETEN  = 0b00110000
        DATE     = 0b00111111

    class MTH(IntFlag):
        MTHONE0 = 0b00000001
        MTHONE1 = 0b00000010
        MTHONE2 = 0b00000100
        MTHTEN0 = 0b00001000
        LPYR    = 0b00010000
        MTHONE  = 0b00001111
        MTHTEN  = 0b00010000
        MTH     = 0b00011111

    class YEAR(IntFlag):
        YRONE0 = 0b00000001
        YRONE1 = 0b00000010
        YRONE2 = 0b00000100
        YRTEN0 = 0b00001000
        YRTEN1 = 0b00010000
        YRTEN2 = 0b00010000
        YRTEN3 = 0b00100000
        YRONE  = 0b00001111
        YRTEN  = 0b11110000
        YR     = 0b11111111

    class CONTROL(IntFlag):
        SQWFS0  = 0b00000001
        SQWFS1  = 0b00000010
        CRSTRIM = 0b00000100
        EXTOSC  = 0b00001000
        ALM0EN  = 0b00010000
        ALM1EN  = 0b00100000
        SQWEN   = 0b01000000
        SQWFS   = 0b00000011

    class OSCTRIM(IntFlag):
        TRIMVAL0 = 0b00000001
        TRIMVAL1 = 0b00000010
        TRIMVAL2 = 0b00000100
        TRIMVAL3 = 0b00001000
        TRIMVAL4 = 0b00010000
        TRIMVAL5 = 0b00100000
        TRIMVAL6 = 0b01000000
        TRIMVAL7 = 0b10000000
        TRIMVAL  = 0b11111111

    class _frame(NamedTuple):
        command: int
        data: list[int]

    class _calc_arg(NamedTuple):
        value: int
        flags: int

    class State(NamedTuple):
        mode: int
        running: bool
        digit: int
        value: list[int]

    class Alarm(NamedTuple):
        enabled: bool
        hour: Union[int, None]
        minute: Union[int, None]

    class Version(NamedTuple):
        major: int
        minor: int

    def __init__(
            self,
            port: str,
            open: bool = False,
            get_pinout: bool = False
    ):
        self._serial = serial.Serial(
            port=None,
            baudrate=9600,
            bytesize=serial.EIGHTBITS,
            parity=serial.PARITY_NONE,
            stopbits=serial.STOPBITS_ONE,
            timeout=self.TIMEOUT,
            xonxoff=False,
            rtscts=False,
            dsrdtr=False,
            write_timeout=self.TIMEOUT,
            inter_byte_timeout=None,
            # exclusive=False
        )
        self._serial.port = port
        self._event: int = self.Event.NONE
        self._pinout: list[int] = None

        if open:
            self.open(get_pinout)

    def __del__(self):
        self.close()

    def open(
            self,
            get_pinout: bool = False
    ):
        if not (self._serial.is_open):
            self._serial.open()
        if get_pinout:
            self.get_pinout()

    def close(self):
        if self._serial.is_open:
            self._serial.close()

    @staticmethod
    def _enum(
            enum_type: IntEnum.__class__,
            value: int
    ) -> Union[Enum, int]:
        value = int(value)
        if value in enum_type.__members__.values():
            return enum_type(value)
        else:
            return value

    @staticmethod
    def _join_u8(
            h: int,
            l: int
    ) -> int:
        return h << 4 | l

    @staticmethod
    def _split_u8(
            x: int
    ) -> tuple[int, int]:
        h = x >> 4
        l = x & 0x0f
        return h, l

    @staticmethod
    def _join_u16(
            h: int,
            l: int
    ) -> int:
        return h << 8 | l

    @staticmethod
    def _split_u16(
            x: int
    ) -> tuple[int, int]:
        h = x >> 8
        l = x & 0xFF
        return h, l

    @classmethod
    def _encode_BCD_u8(
            cls,
            x: int
    ) -> int:
        p, q = divmod(x, 10)
        y = cls._join_u8(p, q)
        return y

    @classmethod
    def _decode_BCD_u8(
            cls,
            x: int
    ) -> int:
        p, q = cls._split_u8(x)
        y = p*10+q
        return y

    @classmethod
    def _encode_BCD_u16(
        cls,
        x: int
    ) -> int:
        p, q = divmod(x, 100)
        y = cls._join_u16(cls._encode_BCD_u8(p), cls._encode_BCD_u8(q))
        return y

    @classmethod
    def _decode_BCD_u16(
        cls,
        x: int
    ) -> int:
        p, q = cls._split_u16(x)
        y = (cls._decode_BCD_u8(p) * 100) + cls._decode_BCD_u8(q)
        return y

    @classmethod
    def _encode_calc_arg(
            cls,
            x: float,
            arg: int = 0
    ) -> _calc_arg:
        if arg == 1:
            negative = cls.CalculationFlag.NEGATIVE1
            dec_point = cls.CalculationFlag.DEC_POINT1
        elif arg == 2:
            negative = cls.CalculationFlag.NEGATIVE2
            dec_point = cls.CalculationFlag.DEC_POINT2
        else:
            negative = cls.CalculationFlag.NEGATIVE
            dec_point = cls.CalculationFlag.DEC_POINT

        flags = 0x00
        if x < 0:
            flags |= negative
            x = -1.0 * x
            limit = 999.5
        else:
            limit = 9999.5

        if x >= limit:
            raise ValueError("Number out of range")

        xf = x * 100.0
        if xf >= limit:
            y = x
        else:
            y = xf
            flags |= dec_point

        return cls._calc_arg(
            value=cls._encode_BCD_u16(int(y+0.5)),
            flags=flags
        )

    @classmethod
    def _decode_calc_arg(
            cls,
            x: int,
            flags: int,
            arg: int = 0
    ) -> float:
        if arg == 1:
            negative = cls.CalculationFlag.NEGATIVE1
            dec_point = cls.CalculationFlag.DEC_POINT1
        elif arg == 2:
            negative = cls.CalculationFlag.NEGATIVE2
            dec_point = cls.CalculationFlag.DEC_POINT2
        else:
            negative = cls.CalculationFlag.NEGATIVE
            dec_point = cls.CalculationFlag.DEC_POINT

        y = float(cls._decode_BCD_u16(x))

        if flags & negative:
            y *= -1.0
        if flags & dec_point:
            y *= 0.01

        return y

    def _read(
            self,
            length: int
    ) -> list[int]:
        buffer = []
        while len(buffer) < length:
            r = self._serial.read(1)
            if len(r) == 0:
                break
            inbyte = r[0]
            if inbyte == self.BUZZER:
                continue
            if (inbyte in self.GetCommand) or (inbyte in self.SetCommand):
                buffer = []

            buffer.append(inbyte)

            if inbyte in self.EndFrame:
                if len(buffer) == 1:
                    buffer = []
                    continue
                else:
                    break

        return buffer

    def _write(
            self,
            buffer: list[int]
    ) -> int:
        length = len(buffer)
        buffer =  bytes(buffer)

        while len(buffer) > 0:
            r = self._serial.write(buffer)
            buffer = buffer[r:]
            # break if ???

        return length-len(buffer)

    def _get_frame(
            self,
            length: int
    ) -> _frame:
        length_b = 2 * length + 2

        r = self._read(max(length_b, self.BUFFER_SIZE))

        if (len(r) == 0) and (length != 0):
            raise serial.SerialTimeoutException('No response from STAF.')
        if len(r) < length_b:
            raise serial.SerialException('Incomplete response from STAF.')
        if (r[length_b-1] not in self.EndFrame) or ((r[0] not in self.GetCommand) and (r[0] not in self.SetCommand)):
            raise serial.SerialException('Invalid response from STAF.')

        data = []
        j = 1
        for i in range(length):
            b = 0
            for k in range(2):
                if ord('0') <= r[j+k] <= ord('9'):
                    h = r[j+k] - ord('0')
                elif ord('A') <= r[j+k] <= ord('F'):
                    h = r[j+k] - ord('A') + 10
                elif ord('a') <= r[j+k] <= ord('f'):
                    h = r[j+k] - ord('a') + 10
                else:
                    raise serial.SerialException('Invalid response from STAF.')
                if k == 0:
                    b = h << 4
                else:
                    data.append(b | h)
                    j += 2

        return self._frame(
            command=r[0],
            data=data
        )

    def _send_frame(
            self,
            command: int,
            data: list[int] = []
    ) -> None:
        if (command not in self.GetCommand) and (command not in self.SetCommand):
            raise ValueError('Invalid command')
        buffer = [command]
        for b in data:
            h = 0
            for k in range(2):
                if k == 0:
                    h = b >> 4
                else:
                    h = b & 0x0F
                if h < 10:
                    buffer.append(h + ord('0'))
                else:
                    buffer.append(h + ord('a') - 10)
        buffer.append(ord('\n'))

        r = self._write(buffer)
        if r < len(buffer):
            raise serial.SerialException(('Nothing' if r == 0 else 'Incomplete frame') + ' sent to STAF.')

    def _check_event(self) -> None:
        while self._serial.in_waiting > 1:  # don't get fooled by a single '\n'
            try:
                command, data = self._get_frame(length=1)
                if command == self.Command.GET_EVENT:
                    self._event |= data[0]
                else:
                    # raise serial.SerialException('Invalid data from STAF')
                    pass
            except serial.SerialTimeoutException:
                pass

    def _execute_command(
            self,
            command: int,
            data_out: list[int] = [],
            length_in: int = 0
    ) -> list[int]:
        self._check_event()
        self._serial.flushInput()

        self._send_frame(
            command=command,
            data=data_out
        )
        if length_in > 0:
            command_in, data_in = self._get_frame(length=length_in)
            if command_in != command:
                raise serial.SerialException('Invalid response from STAF.')
            return data_in
        else:
            return []

    def get_state(self) -> State:
        mode = self.get_mode()
        running = self.is_running()
        digit = self.get_digit()
        value = self.get_value()
        return self.State(
            mode=mode,
            running=running,
            digit=digit,
            value=value
        )

    def set_state(
            self,
            state: State
    ):
        self.set_mode(state.mode)
        self.set_running(state.running)
        self.set_digit(state.digit)
        self.set_value(state.value)

    def get_mode(self) -> Union[int, Mode]:
        r = self._execute_command(
            command=self.Command.GET_MODE,
            length_in=1
        )
        return self._enum(
            enum_type=self.Mode,
            value=r[0]
        )

    def set_mode(
            self,
            mode: Union[int, Mode]
    ) -> None:
        if not (0x00 <= mode <= 0xFF):
            raise ValueError('Invalid mode.')
        self._execute_command(
            command=self.Command.GOTO_MODE,
            data_out=[mode]
        )

    def subscribe_event(
            self,
            event: Union[int, Event]
    ) -> None:
        if not (0x00 <= event <= 0xFF):
            raise ValueError('Invalid event.')
        r = self._execute_command(
            command=self.Command.GET_EVENT,
            data_out=[event],
            length_in=1
        )
        self._event |= r[0]

    def get_event(self) -> Union[int, Event]:
        event = self._event
        self._event = self.Event.NONE
        return event

    def trigger_event(
        self,
        event: Union[int, Event]
    ) -> None:
        if not (0x00 <= event <= 0xFF):
            raise ValueError('Invalid event.')
        self._execute_command(
            command=self.Command.TRIGGER_EVENT,
            data_out=[event],
        )

    def get_pinout(
            self
    ) -> list[int]:
        r = self._execute_command(
            command=self.Command.GET_PINOUT,
            length_in=4
        )
        self._pinout = []
        for b in r:
            self._pinout.append(0xFF & (~(1 << (b >> 4))))
            self._pinout.append(0xFF & (~(1 << (b & 0x0f))))
        return self._pinout

    def get_display(
            self,
            pinout: list[int] = None
    ) -> Union[int, Display]:
        if pinout is None:
            pinout = self._pinout
        if pinout is None:
            raise RuntimeError('Unknown display pinout.')
        if len(pinout) != 8:
            raise ValueError('Invalid display pinout.')
        r = self._execute_command(
            command=self.Command.GET_DISPLAY,
            length_in=4
        )
        display = self.Display.none
        for i, b in enumerate(r):
            for j, p in enumerate(pinout):
                if (b | p) != 0xFF:
                    display |= 1 << ((i << 3) + j)
        return display

    def set_display(
            self,
            display: int,
            pinout: list[int] = None
    ) -> None:
        if pinout is None:
            pinout = self._pinout
        if pinout is None:
            raise RuntimeError('Unknown display pinout.')
        if len(pinout) != 8:
            raise ValueError('Invalid display pinout.')
        data = []
        for i in range(4):
            d = 0xFF;
            for j, p in enumerate(pinout):
                if display & (1 << ((i << 3) +j)):
                    d &= p
            data.append(d)
        self._execute_command(
            command=self.Command.SET_DISPLAY,
            data_out=data
        )

    def free_display(self) -> None:
        self._execute_command(
            command=self.Command.FREE_DISPLAY
        )

    def get_time(
            self,
            tzinfo: timezone = None
    ) -> datetime:
        r = self._execute_command(
            command=self.Command.GET_TIME,
            length_in=8
        )
        return datetime(
            year=self._decode_BCD_u16(self._join_u16(h=r[7], l=r[6])),
            month=self._decode_BCD_u8(r[5]),
            day=self._decode_BCD_u8(r[4]),
            hour=self._decode_BCD_u8(r[3]),
            minute=self._decode_BCD_u8(r[2]),
            second=self._decode_BCD_u8(r[1]),
            microsecond=self._decode_BCD_u8(r[0])*10000,
            tzinfo=tzinfo
        )

    def set_time(
            self,
            hour: int,
            minute: int,
            second: int = None,
            restore_state: bool = False
    ) -> None:
        if not ((0 <= hour <= 23) and (0 <= minute <= 59)):
            raise ValueError('Invalid time setting.')
        if (second is not None) and not (0 <= second <= 59):
            raise ValueError('Invalid time setting.')
        state = None
        if restore_state:
            state = self.get_state()
        self._execute_command(
            command=self.Command.SET_TIME,
            data_out=[
                self._encode_BCD_u8(minute),
                self._encode_BCD_u8(hour)
            ]
        )
        if second is not None:
            v = self.get_debug_value(
                address=self.Register.RTCSEC
            )
            v = (v & self.SEC.ST) | (second & self.SEC.SEC)
            self.set_debug_value(
                address=self.Register.RTCSEC,
                value=v
            )
        if restore_state:
            self.set_state(state)

    def set_date(
            self,
            month: int,
            day: int,
            restore_state: bool = False
    ) -> None:
        d = 29 if (month == 2) else 31 if (((month & 0b1000) >> 3) ^ (month & 0b0001)) else 30
        if not ((1 <= month <= 12) and (1 <= day <= d)):
            raise ValueError('Invalid time setting.')
        state = None
        if restore_state:
            state = self.get_state()
        self._execute_command(
            command=self.Command.SET_DATE,
            data_out=[
                self._encode_BCD_u8(day),
                self._encode_BCD_u8(month)
            ]
        )
        if restore_state:
            self.set_state(state)

    def set_year(
            self,
            year: int,
            restore_state: bool = False
    ) -> None:
        if not (0 <= year <= 9999):
            raise ValueError('Invalid time setting.')
        yh, yl = self._split_u16(self._encode_BCD_u16(year))
        state = None
        if restore_state:
            state = self.get_state()
        self._execute_command(
            command=self.Command.SET_YEAR,
            data_out=[
                yl,
                yh
            ]
        )
        if restore_state:
            self.set_state(state)

    def set_full_time(
            self,
            time: datetime,
            set_seconds: bool = False,
            restore_state: bool = False
    ) -> None:
        state = None
        if restore_state:
            state = self.get_state()
        self.set_time(
            hour = time.hour,
            minute = time.minute,
            second = time.second if set_seconds else None,
        )
        self.set_date(
            month=time.month,
            day=time.day
        )
        self.set_year(
            year=time.year
        )
        if restore_state:
            self.set_state(state)

    def get_alarm(self) -> Alarm:
        r = self._execute_command(
            command=self.Command.GET_ALARM,
            length_in=2
        )
        hour = r[1]
        minute = r[0]
        enabled = (hour != 0xFF) or (minute != 0xFF)
        return self.Alarm(
            enabled=enabled,
            hour=self._decode_BCD_u8(hour) if enabled else None,
            minute=self._decode_BCD_u8(minute) if enabled else None,
        )

    def set_alarm(
            self,
            hour: int,
            minute: int,
            restore_state: bool = False
    ) -> None:
        if not ((0 <= hour <= 23) and (0 <= minute <= 59)):
            raise ValueError('Invalid alarm setting.')
        state = None
        if restore_state:
            state = self.get_state()
        self._execute_command(
            command=self.Command.SET_ALARM,
            data_out=[
                self._encode_BCD_u8(minute),
                self._encode_BCD_u8(hour)
            ]
        )
        if restore_state:
            self.set_state(state)

    def clear_alarm(
            self,
            restore_state: bool = False
    ) -> None:
        state = None
        if restore_state:
            state = self.get_state()
        self._execute_command(
            command=self.Command.SET_ALARM,
            data_out=[
                0xFF,
                0xFF
            ]
        )
        if restore_state:
            self.set_state(state)

    def set_countdown(
            self,
            minute: int,
            second: int,
            restore_state: bool = False
    ) -> None:
        if not ((0 <= minute <= 99) and (0 <= second <= 59)):
            raise ValueError('Invalid countdown setting.')
        state = None
        if restore_state:
            state = self.get_state()
        self._execute_command(
            command=self.Command.SET_COUNTDOWN,
            data_out=[
                self._encode_BCD_u8(second),
                self._encode_BCD_u8(minute)
            ]
        )
        if restore_state:
            self.set_state(state)

    def is_running(self) -> bool:
        r = self._execute_command(
            command=self.Command.GET_RUN_MODE,
            length_in=1
        )
        return bool(r[0])

    def set_running(
            self,
            running: bool
    ) -> None:
        self._execute_command(
            command=self.Command.SET_RUN_MODE,
            data_out=[
                1 if running else 0
            ]
        )

    def get_version(self) -> Version:
        r = self._execute_command(
            command=self.Command.GET_VERSION,
            length_in=2
        )
        return self.Version(
            major=self._decode_BCD_u8(r[1]),
            minor=self._decode_BCD_u8(r[0])
        )

    def get_digit(self) -> int:
        r = self._execute_command(
            command=self.Command.GET_DIGIT,
            length_in=1
        )
        return r[0]

    def set_digit(
            self,
            digit: int
    ) -> None:
        if not (0 <= digit <= 4):
            raise ValueError('Invalid digit.')
        self._execute_command(
            command=self.Command.SELECT_DIGIT,
            data_out=[
                digit
            ]
        )

    def get_digit_value(self) -> int:
        digit = self.get_digit()
        value = self.get_value()
        value = value[digit >> 1]
        if digit & 0x01:
            return value >> 4
        else:
            return value & 0x0f

    def set_digit_value(
            self,
            value: int
    ) -> None:
        if not (0x00 <= value <= 0x0f):
            raise ValueError('Invalid digit value.')
        self._execute_command(
            command=self.Command.SET_DIGIT_VALUE,
            data_out=[
                value
            ]
        )

    def get_value(self) -> list[int]:
        return self._execute_command(
            command=self.Command.GET_VALUE,
            length_in=3
        )

    def set_value(
            self,
            value: list[int]
    ) -> None:
        if len(value) != 3:
            raise ValueError('Invalid value length.')
        for v in value:
            if not (0x00 <= v <= 0xFF):
                raise ValueError('Invalid value.')
        self._execute_command(
            command=self.Command.SET_VALUE,
            data_out=value
        )

    def confirm(self) -> None:
        self._execute_command(
            command=self.Command.CONFIRM
        )

    def set_debug_address(
            self,
            address: int
    ) -> None:
        if not (0x00 <= address <= 0xFF):
            raise ValueError('Invalid address.')
        self._execute_command(
            command=self.Command.SET_DEBUG_ADDRESS,
            data_out=[
                address
            ]
        )

    def get_debug_value(
            self,
            address: int,
            restore_state: bool = False
    ) -> int:
        # if not (0x00 <= address <= 0xFF):
        #     raise ValueError('Invalid address.')
        state = None
        if restore_state:
            state = self.get_state()
        self.set_debug_address(address)
        self.trigger_event(self.Event.TICK)
        value = self.get_value()[0]
        if restore_state:
            self.set_state(state)
        return value

    def set_debug_value(
            self,
            address: int,
            value: int,
            restore_state: bool = False
    ) -> None:
        if not (0x00 <= address <= 0xFF):
            raise ValueError('Invalid address.')
        if not (0x00 <= value <= 0xFF):
            raise ValueError('Invalid value.')
        state = None
        if restore_state:
            state = self.get_state()
        self._execute_command(
            command=self.Command.SET_DEBUG_VALUE,
            data_out=[
                value,
                address
            ]
        )
        if restore_state:
            self.set_state(state)

    def read_memory(
            self,
            address: int
    ) -> int:
        if not (0x0000 <= address <= 0xFFFF):
            raise ValueError('Invalid address.')
        ah, al = self._split_u16(address)
        return self._execute_command(
            command=self.Command.READ_MEMORY,
            data_out=[
                al,
                ah
            ],
            length_in=1
        )[0]

    def write_memory(
            self,
            address: int,
            value: int
    ) -> None:
        if not (0x0000 <= address <= 0xFFFF):
            raise ValueError('Invalid address.')
        if not (0x00 <= value <= 0xFF):
            raise ValueError('Invalid value.')
        ah, al = self._split_u16(address)
        self._execute_command(
            command=self.Command.WRITE_MEMORY,
            data_out=[
                al,
                ah,
                value
            ],
        )

    def get_calibration(
                self,
                restore_state: bool = False
        ) -> int:
        state = None
        if restore_state:
            state = self.get_state()
        self.set_mode(s.Mode.CALIBRATION)
        value = s.get_value()
        calibration = (0 - value[0]) if value[1] else value[0]
        if restore_state:
            self.set_state(state)
        return calibration

    def set_calibration(
            self,
            calibration: int,
            restore_state: bool = False
    ) -> None:
        if not (-0xFF <= calibration <= 0xFF):
            raise ValueError('Invalid calibration value.')
        if calibration < 0:
            sign = 1
            value = 0 - calibration
        else:
            sign = 0
            value = calibration
        state = None
        if restore_state:
            state = self.get_state()
        self._execute_command(
            command=self.Command.SET_CALIBRATION,
            data_out=[
                value,
                sign
            ]
        )
        if restore_state:
            self.set_state(state)

    def set_buzzer_mode(
            self,
            mode: Union[int, BuzzerMode]
    ) -> None:
        if not (mode in list(self.BuzzerMode)):
            raise ValueError('Invalid buzzer mode.')
        self._execute_command(
            command=self.Command.SET_BUZZER_MODE,
            data_out=[
                mode
            ]
        )

    def calculate(
            self,
            a: float,
            b: float,
            operation: Union[int, CalculationFlag],
            restore_state: bool = False
    ) -> float:
        if operation not in list(self.Calculation):
            raise ValueError('Invalid operation.')
        a = self._encode_calc_arg(a, arg=1)
        b = self._encode_calc_arg(b, arg=2)
        ah, al = self._split_u16(a.value)
        bh, bl = self._split_u16(b.value)
        flags = operation | a.flags | b.flags
        state = None
        if restore_state:
            state = self.get_state()
        self._execute_command(
            command=self.Command.CALCULATE,
            data_out=[
                bl,
                bh,
                al,
                ah,
                flags
            ]
        )
        r = self._execute_command(
            command=self.Command.GET_CALCULATION,
            length_in=8
        )
        if r[7]:
            y = float('NaN')
        else:
            y = self._decode_calc_arg(
                x=self._join_u16(
                    h=r[1],
                    l=r[0]
                ),
                flags=r[6]
            )
        if restore_state:
            self.set_state(state)
        return y

    def reset(self):
        self.set_mode(self.Mode.RESET)


# the test also works as usage example:
if __name__ == '__main__':
    s = STAF('com5', open=True, get_pinout=True)
    s.set_buzzer_mode(s.BuzzerMode.DISABLED)

    a = s.get_pinout()
    s.get_display()
    s.set_display(0x0000FFFF)
    s.set_display(s.Display.dot_top | s.Display.dot_bottom)
    s.free_display()

    a = s.get_time()
    s.set_time(
        hour=12,
        minute=34
    )
    s.set_time(
        hour=1,
        minute=23,
        second=45
    )
    s.set_date(
        month=1,
        day=24
    )
    s.set_year(
        year=4567
    )
    s.set_full_time(a, restore_state=True)

    s.set_alarm(
        hour=12,
        minute=34,
        restore_state=True
    )
    a = s.get_alarm()
    s.clear_alarm(restore_state=True)
    a = s.get_alarm()

    s.set_countdown(
        minute=4,
        second=33
    )
    s.set_running(True)
    a = s.is_running()
    s.set_running(False)
    a = s.is_running()

    a = s.get_version()

    s.set_mode(s.Mode.SET_COUNTDOWN)
    a = s.get_mode()
    s.set_digit(1)
    a = s.get_digit()
    s.set_digit_value(0x0E)
    a = s.get_digit_value()
    s.confirm()
    s.set_value([0x33, 0x04, 0x00])
    a = s.get_value()

    s.set_debug_address(s.Register.RTCHSEC)
    s.set_debug_value(
        address=s.Register.ALARMFLAG,
        value=s.ALARMFLAG
    )
    a = s.get_debug_value(
        address=s.Register.ALARMFLAG
    )
    s.set_debug_value(
        address=s.Register.ALARMFLAG,
        value=0
    )
    a = s.get_debug_value(s.Register.ALARMFLAG)

    s.write_memory(
        address=0x022F,
        value=3
    )
    a = s.read_memory(
        address=0x022F
    )

    a=s.get_calibration()
    s.set_calibration(a)


    a=s.calculate(
        a=-4.1,
        b=-2.9,
        operation=s.Calculation.DIVIDE
    )
    a=s.calculate(
        a=9999,
        b=-999,
        operation=s.Calculation.SUBTRACT
    )

    s.set_full_time(
        time=datetime.now(tz=None),  #current local time
        set_seconds=True
    )
    s.set_buzzer_mode(s.BuzzerMode.ENABLED)
    s.reset()
    s.close()

pass
